@@ -2,7 +2,7 @@
 
 """
 Because writing GUI is non-trivial (cf. https://0xacab.org/jvoisin/mat2/issues/3),
-we decided to write a Nautilus extensions instead
+we decided to write a Nautilus extension instead
 (cf. https://0xacab.org/jvoisin/mat2/issues/2).
 
 The code is a little bit convoluted because Gtk isn't thread-safe,
@@ -35,8 +35,8 @@ def _remove_metadata(fpath) -> Tuple[bool, Optional[str]]:
         return False, mtype
     return parser.remove_all(), mtype
 
-class ColumnExtension(GObject.GObject, Nautilus.MenuProvider, Nautilus.LocationWidgetProvider):
-    """ This class adds an item to the right-clic menu in Nautilus. """
+class Mat2Extension(GObject.GObject, Nautilus.MenuProvider, Nautilus.LocationWidgetProvider):
+    """ This class adds an item to the right-click menu in Nautilus. """
 
     def __init__(self):
         super().__init__()
@@ -173,7 +173,7 @@ class ColumnExtension(GObject.GObject, Nautilus.MenuProvider, Nautilus.LocationW
         if fname is None:
             self.infobar_hbox.destroy()
             self.infobar.hide()
-            if len(self.failed_items):
+            if self.failed_items:
                 self.__infobar_failure()
             if not processing_queue.empty():
                 print("Something went wrong, the queue isn't empty :/")
@@ -235,7 +235,7 @@ class ColumnExtension(GObject.GObject, Nautilus.MenuProvider, Nautilus.LocationW
             return None
 
         item = Nautilus.MenuItem(
-            name="MAT2::Remove_metadata",
+            name="mat2::Remove_metadata",
             label="Remove metadata",
             tip="Remove metadata"
         )
